/*
 * GTK gui for acsim
 * golem@mit.edu
 * Based to some degree on the original gui.c.
 */

#include <gtk/gtk.h>
#include <math.h>
#include "acsim.h"


int visible_molecule = MOL_WORLD;


GtkWidget *window;
GtkWidget *drawing_area;
GdkPixmap *pixmap;
GdkColormap *colormap;
GdkGC *vat_gc, *bacteria_gc;

gint width = 600, height = 600;


#define WIN_X(x) ((float)((x) * width)  / (float)SIZE_WORLD)
#define WIN_Y(y) ((float)((y) * height) / (float)SIZE_WORLD)


void
setup_colors()
{
  GdkColor red_color, color;
  int i;

  colormap = gtk_widget_get_colormap(drawing_area);

  /* First allocate the red */
  bacteria_gc = gdk_gc_new(drawing_area->window);
  gdk_color_parse("Red", &red_color);
  gdk_colormap_alloc_color(colormap, &red_color, FALSE, TRUE);
  gdk_gc_set_foreground(bacteria_gc, &red_color);

  /* Now allocate at least 16 patch colors,
   * so that people on 256-bit displays won't be utterly hosed.
   * Hopefully.
   */
  vat_gc = gdk_gc_new(drawing_area->window);
  for (i = 0; i < 16; i++) {
    color.red = color.green = i * 0xFFFF / (16-1);
    color.blue = 0;
    gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);
  }
}


void
draw_world(GdkDrawable *drawable)
{
  int x, y;
  float scale, value;
  GdkColor color;

  scale = get_world_molecular_scale(visible_molecule);

  for (x = 0; x < SIZE_WORLD; x++) {
    for (y = 0; y < SIZE_WORLD; y++) {
      value = get_world_molecular_concentration(visible_molecule, x, y);
      /* This doesn't actually scale it to [0,1] unless min=0.  Ehn. */
      value *= scale;
      color.red = color.green = value * 0xFFFF;
      color.blue = 0;
      gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);
      gdk_gc_set_foreground(vat_gc, &color);
      gdk_draw_rectangle(drawable, vat_gc, TRUE,
                         WIN_X(x), WIN_Y(y),
                         WIN_X(x+1)-WIN_X(x)+1, WIN_Y(y+1)-WIN_Y(y)+1);
    }
  }
}


void
draw_bacteria(GdkDrawable *drawable,
              int          b)
{
  struct coord loc;
  float dir;
  GdkPoint tri[3];

  loc = get_bacteria_loc(b);
  dir = get_bacteria_dir(b);

  tri[0].x = WIN_X( loc.x + 2*cos(dir) );
  tri[0].y = WIN_Y( loc.y + 2*sin(dir) );
  tri[1].x = WIN_X( loc.x + 2*cos(dir + M_PI_2)/4 );
  tri[1].y = WIN_Y( loc.y + 2*sin(dir + M_PI_2)/4 );
  tri[2].x = WIN_X( loc.x + 2*cos(dir - M_PI_2)/4 );
  tri[2].y = WIN_Y( loc.y + 2*sin(dir - M_PI_2)/4 );

  gdk_draw_polygon(drawable, bacteria_gc, TRUE, tri, 3);
}


void
redraw()
{
  GdkRectangle rect;
  rect.x = rect.y = 0;
  rect.width = width;
  rect.height = height;
  gtk_widget_draw(drawing_area, &rect);
}


gint
world_timestep(gpointer data)
{
  int number_of_bacteria, i;

  if (vat_gc == NULL || bacteria_gc == NULL)
    setup_colors();

  /* Clear the drawing area. */
  /*gdk_draw_rectangle(drawing_area->window,
                     drawing_area->style->black_gc,
                     TRUE, 0, 0, width, height);*/


  /* Draw the molecular concentrations in the world. */
  draw_world( pixmap );

  /* Draw the bacteria. */
  number_of_bacteria = get_bacteria_count();
  for (i = 0; i < number_of_bacteria; i++) {
    draw_bacteria( pixmap, i );
  }

  /* Redraw the window. */
  redraw();

  /* Advance the world one step. */
  diffuse_world();
  //evaporate_world();

  for (i = 0; i < number_of_bacteria; i++) {
    bacteria_sensor(i);
    bacteria_reactor(i);
    bacteria_locomotion(i);
  }

  return TRUE;
}


gboolean
expose_event(GtkWidget      *widget,
             GdkEventExpose *event)
{
  /* Draw the pixmap to the window. */
  gdk_draw_pixmap( widget->window,
                   widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                   pixmap,
                   event->area.x, event->area.y,
                   event->area.x, event->area.y,
                   event->area.width, event->area.height );

  return FALSE;
}


gboolean
configure_event(GtkWidget         *widget,
                GdkEventConfigure *event,
                gpointer           data)
{
  width  = widget->allocation.width;
  height = widget->allocation.height;

  if (pixmap)
    gdk_pixmap_unref(pixmap);

  pixmap = gdk_pixmap_new(widget->window, width, height, -1);

  return TRUE;
}


int
main(int   argc,
     char *argv[])
{
  gtk_init(&argc, &argv);

  window = gtk_window_new( GTK_WINDOW_TOPLEVEL );
  gtk_window_set_title( GTK_WINDOW(window), "6.836 ACSim" );

  gtk_signal_connect( GTK_OBJECT(window), "delete-event",
                      gtk_main_quit, NULL );

  drawing_area = gtk_drawing_area_new();
  gtk_container_add( GTK_CONTAINER(window), drawing_area );
  gtk_widget_show( drawing_area );

  gtk_signal_connect( GTK_OBJECT(drawing_area), "expose-event",
                      GTK_SIGNAL_FUNC(expose_event), NULL );
  gtk_signal_connect( GTK_OBJECT(drawing_area), "configure-event",
                      GTK_SIGNAL_FUNC(configure_event), NULL );
  gtk_widget_set_events( drawing_area, GDK_EXPOSURE_MASK );

  gtk_window_set_default_size( GTK_WINDOW(window), width, height );
  gtk_widget_show( window );

  gtk_idle_add( world_timestep, NULL );

  create_world();

  gtk_main();
  return 0;
}
